"""Module for computing cocktail lists of optimal spikes."""

import numpy as np
import csv
from .isodata import IsoData, default_data
from .optimal import optimalspike

def cocktail(type_ = 'pure', filename = 'cookbook.csv', isodatas = None): 
    """Generate double spike cocktail lists.
    
    Args:
        type (str): type of spike, 'pure' or 'real'. Real spikes, such as those from
               Oak Ridge National Labs, contain impurities. See IsoData(element).rawspike
               for their assumed compositions. 
        filename (str): file to store output (a CSV file).
        isodatas (list of IsoData): data on elements to use in cocktail list
            e.g. [IsoData['Ca'], IsoData['Fe']]. If None, it calculates for all
            possible elements.
            
    This routine generates an exhaustive list of all possible double spikes for specified
    elements sorted in order of error.

    Example:
        >>> cocktail('real')
        
    See also optimalspike
    """
    if isodatas is None:
        elements = list(default_data.keys())
        isodatas = [IsoData(el) for el in elements]
        
    print('Writing to '+filename)
    f = open(filename, 'w')
    writer = csv.writer(f)
    
    title = 'A double spike cocktail list: '+type_+' spikes'
    writer.writerow([title])
    writer.writerow([])
    
    for isodata in isodatas:
        element = isodata.element
        print(element)
        isodata = IsoData(element)
        isodata.set_errormodel()
        if (type_=='pure') or (type_=='real' and isodata.nrawspikes() > 1):
            os = optimalspike(isodata,type_)
            optspike = os['optspike']
            opterr = os['opterr']
            optisoinv = os['optisoinv']
            optprop = os['optprop']
            optspikeprop = os['optspikeprop']
            optppmperamu = os['optppmperamu']

            optisoinv = isodata.isoindex(optisoinv)
            
            isoname = isodata.isoname()
            
            def iname(i):
                return isoname[i]
            fv = np.vectorize(iname)
            optisonams = fv(optisoinv)
            
            # write output to file
            isohead = ['iso1','iso2','iso3','iso4']
            
            if type_=='pure':
                header = isohead + isoname + ['spike','sample','error','ppmperamu']
                writer.writerow(header)
                
                for i in range(len(opterr)):
                    line = list(optisonams[i,:]) + list(optspike[i,:]) + list([optprop[i],1-optprop[i]])+list([opterr[i]])+list([optppmperamu[i]])
                    writer.writerow(line)
                writer.writerow([])

            else:
                spikename = ['spike' + str(i+1) for i in range(isodata.rawspike.shape[0])]
                
                header = isohead + isoname + spikename + ['spike','sample','error','ppmperamu']
                writer.writerow(header)
                
                for i in range(len(opterr)):
                    line = list(optisonams[i,:]) + list(optspike[i,:]) + list(optspikeprop[i,:]) + list([optprop[i],1-optprop[i]])+list([opterr[i]])+list([optppmperamu[i]])
                    writer.writerow(line)
                writer.writerow([])
    
    f.close()
    print('Output written to '+filename)
    
if __name__=="__main__":
    #cocktail('pure', filename='cocktail_pure.csv', isodatas=[IsoData('Fe'),IsoData('Ca')])
    cocktail('real', filename='cocktail_real.csv', isodatas=[IsoData('Fe'),IsoData('Ca')])



"""Routines for Monte Carlo mass spec simulation."""
import numpy as np
from numpy.random import default_rng
from .isodata import IsoData
from .inversion import dsinversion
    
def monterun(isodata, prop = None, spike = None, alpha = 0.0, beta = 0.0, n = 1000): 
    """Generate a fake mass spectrometer run by Monte-Carlo simulation.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        prop (float/array): proportion of spike in double spike-sample mix.
            A vector of values can be specified if desired, to reflect changes over a run.
        spike (array): a composition vector for the spike. e.g. [0, 0, 0.5, 0.5] is a 50-50
            mix of 57Fe and 58Fe. If None this is read from isodata.
        alpha (float/array): the natural fractionations (can be float or array of values)
        beta (float/array): the instrumental fractionations (can be float or array of values)
        n (int): number of Monte-Carlo samples to take. Default is 1000.

        Note that behaviour depends on the error model specified in isodata.errormodel. 
    
    Example:
        >>> isodata = IsoData('Fe')
        >>> measured=monterun(isodata,0.5,[0, 0, 0.5, 0.5])
    """
    # Get spike from isodata if not supplied as argument
    if spike is None:
        if isodata.spike is None:
            raise Exception("No spike given")
        else:
            spike = isodata.spike
    spike = np.array(spike)
            
    if isodata.errormodel == {}:
        raise Exception("Need to set errormodel")
    
    standard = isodata.standard
    mass = isodata.mass
    emodel = isodata.errormodel
    spike = spike/sum(spike)
    nisos = isodata.nisos()
    
    if isinstance(alpha, float):
        alpha = alpha * np.ones(n)
    if isinstance(beta, float):
        beta = beta * np.ones(n)
    if isinstance(prop, float):
        prop = prop * np.ones(n)
     
    P = np.log(mass)[np.newaxis,:]
    alpha = alpha[:,np.newaxis]
    beta = beta[:,np.newaxis]
    prop = prop[:,np.newaxis]
    
    standard = standard[np.newaxis,:]
    spike = spike[np.newaxis,:]
    
    
    sample = standard * np.exp( - P * alpha)
    sample = normalise_composition(sample)
    
    mixture = prop * spike + (1-prop) * sample
    
    measured = mixture * np.exp( + P*beta)
    measured = normalise_composition(measured)
    
    def calc_var(data, emod):
        datai = data *  emod['intensity']    
        if emod['type'] == 'fixed-sample':
            datai = datai / (1.0 - prop)
        dataivar = emod['a'] + emod['b'] * datai + emod['c'] * (datai **2)
        return datai, dataivar
    
    measuredi, measuredivar = calc_var(measured, emodel['measured'])
    standardi, standardivar = calc_var(standard, emodel['standard'])
    spikei, spikeivar = calc_var(spike, emodel['spike'])

    rng = default_rng()
    measuredv = rng.normal(loc = measuredi, scale = np.sqrt(measuredivar))
    standardv = None
    spikev = None

    if np.any(standardivar>0.0):
        standardi = np.tile(standardi, (n,1))
        standardivar = np.tile(standardivar, (n,1))
        standardv = rng.normal(loc = standardi, scale = np.sqrt(standardivar))

    if np.any(spikeivar>0.0):
        spikei = np.tile(spikei, (n,1))
        spikeivar = np.tile(spikeivar, (n,1))
        spikev = rng.normal(loc = spikei, scale = np.sqrt(spikeivar))
    
    if standardv is None and spikev is None:
        return measuredv
    elif standardv is not None:
        return measuredv, standardv
    else:
        return measuredv, standardv, spikev

    
def normalise_composition(comp):
    """Normalise an array so rows have unit sum, i.e. rows are compositional vectors."""
    return comp / comp.sum(axis=1)[:, np.newaxis]    
    
if __name__=="__main__":
    idat = IsoData('Fe')
    idat.set_spike([0.0, 0.0, 0.5, 0.5])
    idat.set_errormodel()
    measuredv = monterun(idat, prop = 0.5, alpha = -0.2, beta = 1.8, n = 1000)
    print(measuredv[0:10,:])
    
    out = dsinversion(idat, measuredv)
    import matplotlib.pyplot as plt
    plt.plot(out['alpha'])
    plt.ylabel(r'$\alpha$')
    plt.show()



"""Module for performing linear error propagation."""

import numpy as np
from .isodata import realproptoratioprop, ratio

def errorestimate(isodata, prop = None, spike = None, isoinv = None, errorratio = None, alpha = 0.0, beta = 0.0): 
    """Calculate the error in the natural fractionation factor or a chosen ratio by linear error propagation.

    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        prop (float): proportion of double spike in double spike-sample mix.
        spike (array): the isotopic composition of the spike e.g. [0, 0.5, 0, 0.5]
            corresponds to a 50-50 mixture of the 2nd and 4th isotopes
            (56Fe and 58Fe) in the case of Fe.
        isoinv (array): the isotopes used in the inversion, e.g. [54, 56, 57, 58].
        errorratio (array): by default, the error on the natural fractionation
            factor (known as alpha) is given. Instead, the error on a
            particular ratio can be given by setting errorratio. e.g.
            setting errorratio=[58, 56] will give the error on 58Fe/56Fe.
        alpha, beta (floats): there is a small dependance of the error on the fractionation
            factors (instrumental and natural, or alpha and beta). Values of alpha and
            beta can be set here if desired, although the effect on the optimal spikes
            is slight unless the fractionations are very large.
            
        If spike or isoinv are set as None, values from isodata will be used instead.
        
    Returns:
        error (float): the error on the fractionation factor, or the specified ratio.
        ppmperamu (float): the error converted to an approximate ppm per atomic mass unit
        
    Example:
        >>> isodata_fe = IsoData('Fe')
        >>> error, ppmperamu = errorestimate(isodata_fe,0.5,[0,0.5,0,0.5])
    """
    # Get data from isodata if not supplied as arguments
    if spike is None:
        if isodata.spike is None:
            raise Exception("No spike given")
        else:
            spike = isodata.spike

    if isoinv is None:
        if hasattr(isodata, 'isoinv'):
            isoinv = isodata.isoinv
        else:
            raise Exception('No inversion isotopes set')
    
    # Ensure working with numpy arrays
    spike = np.array(spike)
    spike = spike / sum(spike)
    isoinv = np.array(isoinv)
    
    # Convert isotope mass numbers to index numbers
    errorratio = isodata.isoindex(errorratio)
    isoinv = isodata.isoindex(isoinv)
    
    # Choose denominator isotope from largest spike value
    ix = np.argmax(spike[isoinv])
    deno = isoinv[ix]
    nume = isoinv[isoinv != deno]
    isoinv = np.concatenate((np.array([deno]),nume))
    di = isoinv[0]
    
    # Convert compositional vectors to isotopic ratios
    z, AP, An, AT, Am, AN, AM = ratiodata(isodata, di, prop, spike, alpha, beta)
    
    # Convert measured back to a compositional vector    
    measured = isodata.composition(Am, di)
    
    # Calculate the covariance matrices of n, T, and m
    VAn = calcratiocov(isodata.standard,isodata.errormodel['standard'],di,None,prop)
    VAT = calcratiocov(spike,isodata.errormodel['spike'],di,None,prop)
    VAm = calcratiocov(measured,isodata.errormodel['measured'],di,None,prop)
    
    # invrat gives indices of ratios used in inversion
    invrat = isodata.invrat(isoinv)
    
    Vz, VAN, _ = fcerrorpropagation(z,AP,An,AT,Am,VAn,VAT,VAm,invrat)

    # Error to return
    if errorratio is None:
        error = np.sqrt(Vz[1,1])
    else:
        # Now change coordinates to get variance of ratio we're interested in
        newVAN = changedenomcov(AN,VAN,di,errorratio[1])
        
        isonums = np.arange(isodata.nisos())
        newAni = isonums[isonums != errorratio[1]]
        erat = np.where(errorratio[0] == newAni)[0][0]
        error = np.sqrt(newVAN[erat,erat])
    
    if errorratio is None:
        ppmperamu = (1000000.0 * error) / np.mean(isodata.mass)
    else:
        stdratio = isodata.standard[errorratio[0]] / isodata.standard[errorratio[1]]
        massdiff = np.abs(isodata.mass[errorratio[0]] - isodata.mass[errorratio[1]])
        ppmperamu = (1000000.0 * error) / (stdratio * massdiff)
    
    return error, ppmperamu 

def calcratiocov(composition = None, errormodel = None, di = None, isonorm = None, prop = 0.0): 
    """Calculate the covariance matrix of the ratios based on the given error model and composition."""
    # di is the isotope with which to denominator
    # isonorm are the isotopes to use in the normalisation
    # prop is the proportion of spike in the spike-sample mix
    
    if isonorm is None:
        # normalise so that the sum of all beams is the mean intensity
        isonorm = np.arange(composition.shape[0])
    
    # first normalise composition so it is really a composition (unit sum)
    composition = composition / sum(composition)
    
    meanbeams = errormodel['intensity']*composition / sum(composition[isonorm])
    
    if errormodel['type']=='fixed-sample':
        meanbeams = meanbeams / (1.0 - prop)

    covbeams = calcbeamcov(meanbeams,errormodel)
    V = covbeamtoratio(meanbeams,covbeams,di)
    return V
    
def calcbeamcov(meanbeams = None,errormodel = None): 
    """Calculate beam covariance matrix."""
    beamvar = errormodel['a'] + errormodel['b']*meanbeams + errormodel['c']*(meanbeams ** 2)
    return np.diag(beamvar)
    
def covbeamtoratio(meanbeams = None,covbeams = None,di = None): 
    """Convert a covariance matrix for beams to one for ratios."""
    # di is the isotope to denominator with
    # assumes last row and column of M correspond to denominator
    isonums = np.arange(len(meanbeams))
    ni = isonums[isonums != di]
    n = meanbeams[ni]
    d = meanbeams[di]
    ii = np.concatenate((ni, np.array([di]))) # move denominator to end
    M = covbeams[ii, :][:,ii]
    
    D = np.diag(1/d * np.ones(len(n)))
    S = - n / (d ** 2)
    A = np.hstack((D, S[:,np.newaxis]))
    
    V = A @ M @ A.T
    return V

def changedenomcov(data = None, datacov = None, olddi = None, newdi = None): 
    """Change denominator of covariance matrix for given set of ratios."""
    nisos = len(data) + 1
    oldni = np.concatenate((np.arange(olddi),np.arange(olddi+1,nisos)))
    dataplus = np.concatenate((data[0:olddi],np.array([1]),data[olddi:]))  
    
    newni = np.concatenate((np.arange(newdi),np.arange(newdi+1,nisos)))
    
    datacovplus = np.zeros((nisos,nisos))
    #datacovplus[:,oldni][oldni,:] = datacov
    # There must be a vectorized way of doing this!
    for i,n in enumerate(oldni):
        for j,m in enumerate(oldni):
            datacovplus[n,m] = datacov[i,j]

    A = np.eye(nisos) / dataplus[newdi]
    A[:,newdi] = A[:,newdi] - dataplus / (dataplus[newdi] ** 2)
    newdatacovplus = A @ datacovplus @ A.T
    newdatacov = newdatacovplus[:,newni][newni,:]
    return newdatacov

def ratiodata(isodata, di, prop, spike = None, alpha = 0.0, beta = 0.0):
    """Calculate isotopic ratios describing system.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        di (int): denominator isotope, e.g. 56
        prop (float): proportion of double spike in double spike-sample mix.
        spike (array): the isotopic composition of the spike e.g. [0, 0.5, 0, 0.5]
            corresponds to a 50-50 mixture of the 2nd and 4th isotopes
            (56Fe and 58Fe) in the case of Fe.
        alpha (float): natural fractionation factor
        beta (float): instrumental fractionation factor
        
        If spike is set as None, values from isodata will be used instead.
    
    Returns:
        z (array): vector of model parameters (lambda, alpha, beta)
        AP (array): log of ratio of atomic masses
        An (array): isotopic ratios of standard/ unspiked run
        AT (array): isotopic ratios of spike
        Am (array): isotopic ratios of measurement
        AN (array): isotopic ratios of sample
        AM (array): isotopic ratios of mixture
    """
    if spike is None:
        if isodata.spike is None:
            raise Exception("No spike given")
        else:
            spike = isodata.spike
    
    # Calculate ratios
    AP = np.log(isodata.ratio(np.array(isodata.mass), di))
    AT = isodata.ratio(np.array(spike), di)
    An = isodata.ratio(np.array(isodata.standard), di)
    
    # Now calculate sample ratio, lambda etc
    AN = An * np.exp(-AP*alpha)
    lambda_ = realproptoratioprop(prop, AT, AN)
    z = np.array([lambda_,alpha,beta])
    AM = lambda_*AT + (1-lambda_)*AN
    Am = AM * np.exp(AP*beta)
    
    return z, AP, An, AT, Am, AN, AM

def z_sensitivity(z, P, n, T, m):
    """Calculate the derivatives of the z=(lambda,alpha, beta) vector with respect to the input n, T, m ratios."""
    lambda_ = z[0]
    alpha = z[1]
    beta = z[2]
    
    N = n * np.exp(- P * alpha)
    M = m * np.exp(- P * beta)
    
    # calculate various Jacobian matrices
    dfdlambda = T - N*(1 + alpha*P)
    dfdu = -N*P
    dfdbeta = M*P
    dfdy = np.array([dfdlambda,dfdu,dfdbeta]).T
    dfdT = lambda_*np.eye(3)
    dfdm = - np.diag(np.exp(- beta * P))
    dfdn = (1 - lambda_) * np.diag(np.exp(- alpha * P))
    
    ## matrix to convert from (lambda, (1-lambda)alpha,beta) to (lambda,alpha,beta)
    K = np.array([[1,0,0],
                  [(alpha / (1 - lambda_)),(1 / (1 - lambda_)),0],
                  [0,0,1]])
    dzdT = - K @ (np.linalg.solve(dfdy,dfdT))
    dzdm = - K @ (np.linalg.solve(dfdy,dfdm))
    dzdn = - K @ (np.linalg.solve(dfdy,dfdn))
    
    return dzdn, dzdm, dzdT

def sensitivity(z, AP, An, AT, Am, invrat):
    """Returns the partial derivatives describing the sensitivity of model outputs to model inputs.
        
    Args:
        z (array): vector of (lambda, alpha, beta)
        AP (array): log of ratio of atomic masses
        An (array): isotopic ratios of standard/ unspiked run
        AT (array): isotopic ratios of spike
        Am (array): isotopic ratios of measurement
        invrat: indices of isotopic ratios used in inversion
            
    Returns the derivatives:
        dzdAn, dzdAT, dzdAm,
        dANdAn, dANdAT, dANdAm,
        dAMdAn, dAMdAT, dAMdAm,
        
        where AN are isotopic ratios of sample
        and AM are isotopic ratios of mixture.        
    """
    lambda_ = z[0]
    alpha = z[1]
    beta = z[2]
    AM = Am * np.exp(- AP * beta)
    AN = An * np.exp(- AP * alpha)
    
    P = AP[invrat]
    n = An[invrat]
    T = AT[invrat]
    m = Am[invrat]
    dzdn, dzdm, dzdT = z_sensitivity(z, P, n, T, m)
    
    # full matrices for all ratios
    nratios = len(An)
    dzdAT = np.zeros((3,nratios))
    dzdAn = np.zeros((3,nratios))
    dzdAm = np.zeros((3,nratios))
    dzdAT[0:3,:][:,invrat] = dzdT
    dzdAn[0:3,:][:,invrat] = dzdn
    dzdAm[0:3,:][:,invrat] = dzdm
    
    dalphadAT = dzdAT[1,:]
    dalphadAn = dzdAn[1,:]
    dalphadAm = dzdAm[1,:]
    
    NP = AN*AP
    NP = NP[:,np.newaxis]
    dANdAT = - NP @ dalphadAT[np.newaxis,:]
    dANdAn = np.diag(np.exp(- alpha*AP)) - NP @ dalphadAn[np.newaxis,:]
    dANdAm = - NP @ dalphadAm[np.newaxis,:]
    
    dbetadAT = dzdAT[2,:]
    dbetadAn = dzdAn[2,:]
    dbetadAm = dzdAm[2,:]
    
    MP = AM*AP
    MP = MP[:,np.newaxis]
    dAMdAT = - MP @ dbetadAT[np.newaxis,:]
    dAMdAn = - MP @ dbetadAn[np.newaxis,:]
    dAMdAm = np.diag(np.exp(- beta*AP)) - MP @ dbetadAm[np.newaxis,:]
    
    return dzdAn, dzdAT, dzdAm, dANdAn, dANdAT, dANdAm, dAMdAn, dAMdAT, dAMdAm
    
def fcerrorpropagation(z,AP,An,AT,Am,VAn,VAT,VAm,invrat): 
    """Linear error propagation for the fractionation correction."""
    Vn = VAn[invrat,:][:,invrat]
    VT = VAT[invrat,:][:,invrat]
    Vm = VAm[invrat,:][:,invrat]
    
    dzdAn, dzdAT, dzdAm, dANdAn, dANdAT, dANdAm, dAMdAn, dAMdAT, dAMdAm = sensitivity(z, AP, An, AT, Am, invrat)
    
    # Covariance matix for z=(lambda,beta,alpha), sample, mixture
    Vz = dzdAn @ VAn @ dzdAn.T + dzdAT @ VAT @ dzdAT.T + dzdAm @ VAm @ dzdAm.T 
    VAN = dANdAn @ VAn @ dANdAn.T + dANdAT @ VAT @ dANdAT.T + dANdAm @ VAm @ dANdAm.T
    VAM = dAMdAn @ VAn @ dAMdAn.T + dAMdAT @ VAT @ dAMdAT.T + dAMdAm @ VAm @ dAMdAm.T
    
    return Vz, VAN, VAM

if __name__=="__main__":
    from .isodata import IsoData
    isodata = IsoData('Fe')
    isodata.set_spike([0.0, 0.0, 0.5, 0.5])
    alpha_err, ppm_err = errorestimate(isodata, prop = 0.5, alpha = -0.2, beta = 1.8 )
    print(alpha_err, ppm_err)



"""
The double spike toolbox
========================

Error propagation and data reduction for the double spike technique in mass spectrometry.

The workings of this package are described in:

Rudge J.F., Reynolds B.C., Bourdon B. The double spike toolbox (2009) Chem. Geol. 265:420-431
https://dx.doi.org/10.1016/j.chemgeo.2009.05.010

and at:

https://johnrudge.com/doublespike

Functions:
    cocktail               - Generate double spike cocktail lists
    dsinversion            - Do the double spike inversion for a given set of measurements
    errorcurve             - A plot of error as a function of double spike-sample proportions for a given
                             double spike composition
    errorcurve2            - A plot of error as a function of double spike proportions for a given double
                             spike-sample proportion
    errorcurve2d           - A 2D contour plot of error as a function of double spike composition and
                             double spike-sample proportions
    errorcurveoptimalspike - Plot error curves for the optimal spike composition
    errorestimate          - Calculate the error in the natural fractionation factor or a chosen ratio
                             by linear error propagation
    monterun               - Generate a fake mass spectrometer run by Monte-Carlo simulation
    optimalspike           - Find the optimal double spike composition and double spike-sample mixture
                             proportions
"""

__version__ = '0.9.1'
__author__ = 'John F. Rudge'

from .isodata import IsoData
from .inversion import dsinversion
from .monte import monterun
from .cocktail import cocktail
from .errors import errorestimate, ratiodata, sensitivity
from .optimal import optimalspike
from .plotting import errorcurve, errorcurve2, errorcurve2d, errorcurveoptimalspike



"""Module which handles the data on isotope systems."""
import csv
import numpy as np
import pkg_resources

# Fundamental constants
elementarycharge = 1.60217646e-19 # coulombs
k = 1.3806504e-23  # m^2 kg s^-2 K^-1, Boltzmann constant

def loadrawdata(filename = None):
    """Read in isotope system datafile and return a python dictionary with data."""
    if filename is None:
        resource_package = __name__
        resource_path = '/'.join(('data', 'maininput.csv'))  # Do not use os.path.join()
        filename = pkg_resources.resource_filename(resource_package, resource_path)
    
    f = open(filename, 'r')
    csvreader = csv.reader(f)
    
    data = {}

    next(csvreader) # ignore header row
    j = 0
    els = []

    for row in csvreader:
        if not len(row[0])==0:
            el = row[0].strip('"')
            j = j + 1
            els.append(el)
            i = 0
            data[el]={'element':el,'isonum':[], 'mass':[], 'standard':[], 'rawspike':[]}
        data[el]['isonum'].append(int(row[1]))
        data[el]['mass'].append(float(row[2]))
        data[el]['standard'].append(float(row[3]))
        rs = [float(r) for r in row[4:] if len(r)>0]
        data[el]['rawspike'].append(rs)
        
    # Renormalise compositions, convert to numpy format    
    for el in els:
        data[el]['isonum'] = np.array(data[el]['isonum'])
        data[el]['mass'] = np.array(data[el]['mass'])
        
        s = np.array(data[el]['standard'])
        data[el]['standard'] = s / sum(s)

        rs = np.array(data[el]['rawspike'])
        rs = rs/rs.sum(axis=0)
        data[el]['rawspike'] = rs.T
        if len(data[el]['rawspike'])==0:
            data[el]['rawspike'] = None
    
    f.close()
    return data

default_data = loadrawdata()

class IsoData():
    """Object which stores data on an isotope system.
    
    Args:
        element (str): the element string (e.g. 'Fe')
    
    Attributes:
        element (str): the element e.g. 'Fe'
        isonum (array): the isotopes of this element e.g. [54, 56, 57, 58]
        mass (array): the atomic masses
        standard (array): composition of a standard as a vector
        spike (array): composition of a double spike used
        isoinv (array): the 4 isotopes to use the inversion
        rawspike (array): composition of single spikes available
        errormodel (dict): dictionary describing the errormodel
    """
    
    def __init__(self, element):
        """Given an element, initialise with some sensible default values.""" 
        if element in default_data.keys():
            default = default_data[element]
            self.element = default['element']
            self.isonum = default['isonum']
            self.mass = default['mass']
            self.standard = default['standard']
            self.rawspike = default['rawspike']
        else:
            self.element = element
            self.isonum = None
            self.mass = None
            self.standard = None
            self.rawspike = None
        self.spike = None
        if self.nisos() == 4:
            self.isoinv = self.isonum
        else:
            self.isoinv = None
        self.errormodel = {}
        self.set_errormodel()
    
    def __repr__(self):
        return "IsoData()"
    
    def __str__(self):
        txt ='\n '.join("%s: %s" % item for item in vars(self).items())
        return txt
    
    def set_element(self, element):
        self.element = element
        
    def set_mass(self, mass):
        self.mass = np.array(mass)
    
    def set_isonum(self, isonum):
        self.isonum = np.array(isonum, dtype=int)
        
    def set_isoinv(self, isoinv):
        self.isoinv = np.array(isoinv, dtype=int)
    
    def set_standard(self, standard):
        self.standard = np.array(standard)

    def set_spike(self, spike):
        self.spike = np.array(spike)
    
    def set_rawspike(self, rawspike):
        self.rawspike = np.array(rawspike)
    
    def isoindex(self, ix):
        """Give the data index corresponding to a given isotope number e.g. 56->1."""
        
        def isonum_to_idx(k):
            quest = np.where(self.isonum ==k)[0]
            if len(quest)==0:
                return k
            else:
                return quest[0]
        
        if ix is None:
            return None
        elif type(ix) == int:
            return isonum_to_idx(ix)
        elif type(ix) == list:
            return [isonum_to_idx(i) for i in ix]
        else:
            # assume a numpy array
            f = np.vectorize(isonum_to_idx)
            return f(ix)
    
    def isoname(self):
        """Names of the isotopes."""
        return [self.element + str(i) for i in self.isonum]
    
    def isolabel(self):
        """Isotope labels for plotting."""
        return ['$^{' +  str(i) +'}$' + self.element for i in self.isonum]
    
    def rawspikelabel(self):
        """Single spike labels for plotting."""
        return ['spike ' + str(i+1) for i in range(self.rawspike.shape[0])]
    
    def nisos(self):
        """Number of isotopes in system."""
        return len(self.isonum)
    
    def nratios(self):
        """Number of isotope ratios to describe system."""
        return self.nisos() - 1
    
    def nrawspikes(self):
        """Number of single spikes available."""
        if self.rawspike is None:
            return 0
        else:
            return self.rawspike.shape[0]
    
    def set_errormodel(self, intensity = 10.0, deltat = 8.0, R = 1e11, T = 300.0, radiogenic = None, measured_type = 'fixed-total'): 
        """Set the error model used for error estimates and monte carlo runs.
        
        Args:
            intensity (float): Total beam intensity in volts
            delta_t (float): Integration time in seconds
            R (float): resistance in Ohms
            T (float): temperature in K
            radiogenic (str): If True put errors on the standard (unspiked) run, if False do not.
                        If None, then 'Pb','Sr','Hf','Os','Nd' are assumed radiogenic, others not.
            measured_type (str): If 'fixed-total' then total beam intensity of mixture is fixed,
                           if 'fixed-sample' then voltage for the sample is fixed.
        """
        if radiogenic is None:
            if self.element in ['Pb','Sr','Hf','Os','Nd']:
                radiogenic = True
            else:
                radiogenic = False
        
        a = 4 * k * T * R / deltat
        b = elementarycharge * R / deltat
        
        nisos = self.nisos()
        
        # by default assume Johnson noise and counting statistics
        self.errormodel['measured']={
            'type': measured_type,
            'intensity': intensity,
            'a': a*np.ones(nisos),
            'b': b*np.ones(nisos),
            'c': 0.0*np.ones(nisos)
            }
        self.errormodel['spike']={
            'type': 'fixed-total',
            'intensity': intensity,
            'a': 0.0*np.ones(nisos),
            'b': 0.0*np.ones(nisos),
            'c': 0.0*np.ones(nisos)
            }
        if radiogenic:
            # if a radiogenic isotope then put errors on the standard (unspiked) run
            self.errormodel['standard']={
                'type': 'fixed-total',
                'intensity': intensity,
                'a': a*np.ones(nisos),
                'b': b*np.ones(nisos),
                'c': 0.0*np.ones(nisos)
                }
        else:
            # otherwise assume no errors on standard composition
            self.errormodel['standard']={
                'type': 'fixed-total',
                'intensity': intensity,
                'a': 0.0*np.ones(nisos),
                'b': 0.0*np.ones(nisos),
                'c': 0.0*np.ones(nisos)
                }

    def set_custom_errormodel(self, errormodel):
        """Set the error model used for error estimates and monte carlo runs.
        
        Args:
            errormodel: A dictionary giving the complete errormodel.
                    
        See IsoData.errormodel for format of dictionary.
        """
        self.errormodel = errormodel
        
    def ratio(self, composition, denominator_isotope):
        """Convert a compositional array into array of isotopic ratios."""
        di = self.isoindex(denominator_isotope)
        ni = np.arange(self.nisos())
        ni = ni[ni != di]
        return composition[...,ni]/composition[...,di,np.newaxis]
    
    def composition(self, data, denominator_isotope):
        """Convert an array of isotopic ratios to an array of compositional vectors."""
        di = self.isoindex(denominator_isotope)
        ni = np.arange(self.nisos())
        ni = ni[ni != di]
        comp_shape = list(data.shape)
        comp_shape[-1] += 1  
        comp = np.ones(comp_shape)
        comp[ni] = data
        comp = normalise_composition(comp)
        return comp
        
    def invrat(self, isoinv = None):
        """Indices of isotope ratios used in inversion."""
        if isoinv is None:
            isoinv = self.isoinv
            
        isoinv = self.isoindex(isoinv) # convert to indices
        
        isonum = np.arange(self.nisos())
        isonum = isonum[isonum != isoinv[0]]
        isonum = np.concatenate((np.array([isoinv[0]]), isonum))

        invrat = np.array([np.where(isonum == i)[0][0] for i in isoinv])
        invrat = invrat[1:] - 1
        return invrat
    
    def rationame(self, denominator_isotope):
        """Names of the isotopic ratios."""
        di = self.isoindex(denominator_isotope)
        ni = np.arange(self.nisos())
        ni = ni[ni != di]
        return np.array([str(i) + self.element +'/' + str(self.isonum[di]) + self.element   for i in self.isonum[ni]])
    
    def ratioidx(self, numerator_isotope, denominator_isotope):
        """The index corresponding to a particular isotopic ratio."""
        di = self.isoindex(denominator_isotope)
        ni = self.isoindex(numerator_isotope)
        if ni<di:
            return ni
        if ni==di:
            return None
        if ni>di:
            return ni-1
        
def normalise_composition(comp):
    """Normalise rows of an array to unit sum, i.e. rows are compositional vectors."""
    s = comp.sum(axis=-1)
    if type(s) is float:
        return comp/s
    else:
        return comp / s[..., np.newaxis]

def ratioproptorealprop(lambda_, ratio_a, ratio_b): 
    """Convert a proportion in ratio space to one per mole."""
    a = 1 + sum(ratio_a)
    b = 1 + sum(ratio_b)
    return lambda_*a / (lambda_*a+ (1-lambda_)*b)

def realproptoratioprop(prop, ratio_a, ratio_b): 
    """Convert a proportion per mole into ratio space."""
    a = 1 + sum(ratio_a)
    b = 1 + sum(ratio_b)
    return prop*b / (prop*b+ (1-prop)*a) 

def ratio(data, isoidx):
    """Convert data to isotope ratios based on choice of isotopes. First index is the denominator index."""
    di = isoidx[0]
    ni = isoidx[1:]
    return data[...,ni]/data[...,di,np.newaxis]


if __name__=="__main__":
    idat = IsoData('Fe')
    print(idat)




"""Plotting routines using matplotlib."""
import numpy as np
import matplotlib.pyplot as plt
from .errors import errorestimate
from .optimal import optimalspike
    
def errorcurve2d(isodata, type_ = 'pure', isospike = None, isoinv = None, errorratio = None, alpha = 0.0, beta = 0.0, resolution = 100,threshold = 0.25,ncontour = 25,plottype = 'default', **kwargs): 
    """2D contour plot of error as a function of double spike composition and double spike-sample proportions.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        type (str): type of spike, 'pure' or 'real'. Real spikes, such as those from
            Oak Ridge National Labs, contain impurities. See isodata.rawspike for the assumed compositions.
            By default pure spikes are used.
        isospike(array): the isotopes used in the double spike e.g. [54, 57].
        isoinv (array): the isotopes used in the inversion, e.g. [54, 56, 57, 58].
        errorratio (array): by default, the error on the natural fractionation
            factor (known as alpha) is given. Instead, the error on a
            particular ratio can be given by setting errorratio. e.g.
            setting errorratio=[58, 56] will give the error on 58Fe/56Fe.
        alpha, beta (float): there is a small dependance of the error on the fractionation
            factors (natural and instrumental). Values of alpha and
            beta can be set here if desired, although the effect on the optimal spikes
            is slight unless the fractionations are very large. Default is zero.
        resolution (int): number of grid points in x and y. Default is 100.
        threshold (float): maximum contour to plot, relative to the minimum error.
            Default is 0.25 i.e. 25% in excess of the minimum.
        ncontour (int): number of countours. Default is 25.
        plottype (str): by default, the error is plotted. By setting this to 'ppmperamu'
            an estimate of the ppm per amu is plotted instead.
        **kwargs: additional keyword arguments are passed to contour command.
    
    Example:
        >>> isodata_fe = IsoData('Fe')
        >>> errorcurve2d(isodata_fe,'pure',[57, 58])
        
    See also errorestimate
    """
    # Get data from isodata if not supplied as arguments
    if isoinv is None:
        if hasattr(isodata, 'isoinv'):
            isoinv = isodata.isoinv
        else:
            isoinv = isodata.isonum[0:4]

    if isospike is None:
        isospike = np.array([0,1])
    else:
        isodata.isoindex(isospike)
    
    # Ensure working with numpy arrays
    isoinv = np.array(isoinv)
    isospike = np.array(isospike)
    
    # Convert isotope mass numbers to index numbers
    errorratio = isodata.isoindex(errorratio)
    isoinv = isodata.isoindex(isoinv)
    isospike = isodata.isoindex(isospike)

    if type_ == 'pure':
        spike1 = np.zeros(isodata.nisos())
        spike2 = np.zeros(isodata.nisos())
        spike1[isospike[0]] = 1
        spike2[isospike[1]] = 1
    else:
        spike1 = isodata.rawspike[isospike[0],:]
        spike2 = isodata.rawspike[isospike[1],:]
    
    prop = np.linspace(0.001,0.999,resolution)
    spikeprop = np.linspace(0.001,0.999,resolution)
    iv,jv = np.meshgrid(np.arange(len(prop)),np.arange(len(spikeprop)))
    
    def fun(i,j):
        return errorestimate(isodata,prop[i],spikeprop[j]*spike1 + (1 - spikeprop[j])*spike2,isoinv,errorratio,alpha,beta)
    vfun = np.vectorize(fun)
    errvals,ppmperamuvals = vfun(iv, jv)
    
    os = optimalspike(isodata,type_,isospike,isoinv,errorratio,alpha,beta)
    
    if plottype == 'ppmperamu':
        C = plt.contour(prop,spikeprop,ppmperamuvals,np.linspace(os['optppmperamu'],(1 + threshold) * os['optppmperamu'],ncontour + 1),**kwargs)
    else:
        C = plt.contour(prop,spikeprop,errvals,np.linspace(os['opterr'][0],(1 + threshold) * os['opterr'][0],ncontour + 1),**kwargs)
    
    plt.xlim(np.array([0,1]))
    plt.ylim(np.array([0,1]))
    plt.xlabel('proportion of double spike in double spike-sample mix')
    
    isolabel = isodata.isolabel()
    if type_ == 'pure':
        plt.ylabel('proportion of '+isolabel[isospike[0]]+' in '+isolabel[isospike[0]]+'-'+isolabel[isospike[1]]+' double spike')
    else:
        rawspikelabel = isodata.rawspikelabel()
        plt.ylabel('proportion of '+rawspikelabel[isospike[0]]+' in '+rawspikelabel[isospike[0]]+'-'+rawspikelabel[isospike[1]]+' double spike')
    isoinv = isodata.isoindex(isoinv)
    invisostring = isolabel[isoinv[0]]+', '+isolabel[isoinv[1]]+', '+isolabel[isoinv[2]]+', '+isolabel[isoinv[3]]+' inversion'
    if errorratio is None:
        plt.title(r'Error in $\alpha$ (1SD) with '+invisostring)
    else:
        plt.title('Error in '+isolabel[errorratio[0]]+'/',isolabel[errorratio[1]]+' (1SD) with '+invisostring)
    
    plt.plot(os['optprop'][0],os['optspikeprop'][0,isospike[0]],'rx')

def errorcurve(isodata,spike = None,isoinv = None,errorratio = None,alpha = 0.0,beta = 0.0,plottype = 'default',**kwargs): 
    """Plot of error as a function of double spike-sample proportions for a given double spike composition.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        spike (array): the composition of the double spike as a composition vector e.g. [0, 0, 0.5, 0.5]
            represents a 50-50 mixture of the third and fourth isotopes (57-58 for Fe).
        isoinv (array): the isotopes used in the inversion, e.g. [54, 56, 57, 58].
            By default the first four isotopes are chosen.
        errorratio (array): by default, the error on the natural fractionation
            factor (known as alpha) is given. Instead, the error on a
            particular ratio can be given by setting errorratio. e.g.
            setting errorratio=[58, 56] will give the error on 58Fe/56Fe.
        alpha, beta (float): there is a small dependance of the error on the fractionation
            factors (natural and instrumental). Values of beta and
            alpha can be set here if desired, although the effect on the optimal spikes
            is slight unless the fractionations are very large.
        plottype (str): by default, the error is plotted. By setting this to 'ppmperamu'
            an estimate of the ppm per amu is plotted instead.
        **kwargs -- additional keyword arguments are passed to the plot command.
    
    Example:
        >>> isodata_fe = IsoData('Fe')
        >>> errorcurve(isodata_fe,[0, 0, 0.5, 0.5])
    
    See also errorestimate, errorcurve2
    """
    # Get data from isodata if not supplied as arguments
    if isoinv is None:
        if hasattr(isodata, 'isoinv'):
            isoinv = isodata.isoinv
        else:
            isoinv = isodata.isonum[0:4]
    spike = np.array(spike)
    spike = spike / sum(spike)
    # Convert isotope mass numbers to index numbers
    errorratio = isodata.isoindex(errorratio)
    isoinv = isodata.isoindex(isoinv)
    pvals = np.linspace(0.001,0.999,1000)
    errvals = np.zeros(len(pvals))
    ppmperamuvals = np.zeros(len(pvals))
    for i in range(len(pvals)):
        errvals[i],ppmperamuvals[i] = errorestimate(isodata,pvals[i],spike,isoinv,errorratio,alpha,beta)
    
    if plottype=='ppmperamu':
        plotvals = ppmperamuvals
    else:
        plotvals = errvals
    
    plt.plot(pvals,plotvals,**kwargs)
    mine = np.amin(plotvals)
    plt.xlim(np.array([0,1]))
    plt.ylim(np.array([0,5 * mine]))
    plt.xlabel('proportion of double spike in double spike-sample mix')
    isolabel = isodata.isolabel()
    if errorratio is None:
        plt.ylabel(r'Error in $\alpha$ (1SD)')
    else:
        plt.ylabel('Error in '+isodata.isolabel[errorratio[0]]+'/'+isolabel[errorratio[1]]+' (1SD)')
    
    plt.title(isolabel[isoinv[0]]+', '+isolabel[isoinv[1]]+', '+isolabel[isoinv[2]]+', '+isolabel[isoinv[3]]+' inversion')


def errorcurve2(isodata,type_ = 'pure',prop = 0.5,isospike = None, isoinv = None,errorratio = None,alpha = 0.0,beta = 0.0,plottype = 'default',**kwargs): 
    """Plot of error as a function of double spike proportions for a given double spike-sample proportion.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        type (str): type of spike, 'pure' or 'real'. Real spikes, such as those from
            Oak Ridge National Labs, contain impurities. See isodata.rawspike
            for their assumed compositions. By default pure spikes are used.
        prop (float): proportion of double spike in double spike-sample mixture e.g. 0.5.
        isospike (array): the isotopes used in the double spike e.g. [54, 57].
        isoinv (array): the isotopes used in the inversion, e.g. [54, 56, 57, 58].
        errorratio (array):by default, the error on the natural fractionation
            factor (known as alpha or alpha) is given. Instead, the
            error on a particular ratio can be given by setting errorratio. e.g.
            setting errorratio=[58, 56] will give the error on 58Fe/56Fe.
        alpha, beta (float): there is a small dependance of the error on the fractionation
            factors (natural and instrumental). Values of beta and
            alpha can be set here if desired, although the effect on the optimal spikes
            is slight unless the fractionations are very large. 
        plottype (str): by default, the error is plotted. By setting this to 'ppmperamu'
            an estimate of the ppm per amu is plotted instead.
        **kwargs -- additional arguments are passed to the plot command.
    
    Example:
        >>> isodata_fe = IsoData('Fe')
        >>> errorcurve2(isodata_fe,'real',0.5,[54,57])
    
    See also errorestimate, errorcurve
    """
    # Get data from isodata if not supplied as arguments
    if isoinv is None:
        if hasattr(isodata, 'isoinv'):
            isoinv = isodata.isoinv
        else:
            raise Exception('Inversion isotopes not set')
    
    if isospike is None:
        raise Exception('isospike not set')
    
    rawspike = isodata.rawspike
    # Convert isotope mass numbers to index numbers
    errorratio = isodata.isoindex(errorratio)
    isoinv = isodata.isoindex(isoinv)
    isospike = isodata.isoindex(isospike)
    qvals = np.linspace(0.001,0.999,1000)
    errvals = np.zeros(len(qvals))
    ppmperamuvals = np.zeros(len(qvals))
    
    if type_ == "pure":
        spikevector1 = np.zeros(isodata.nisos())
        spikevector1[isospike[0]] = 1.0
        spikevector2 = np.zeros(isodata.nisos())
        spikevector2[isospike[1]] = 1.0
    else:
        spikevector1 = isodata.rawspike[isospike[0], :]
        spikevector2 = isodata.rawspike[isospike[1], :]
    
    for i in range(len(qvals)):
        spike = qvals[i] * spikevector1 + (1-qvals[i]) * spikevector2
        errvals[i],ppmperamuvals[i] = errorestimate(isodata,prop,spike,isoinv,errorratio,alpha,beta)
    
    if plottype=='ppmperamu':
        plotvals = ppmperamuvals
    else:
        plotvals = errvals
    
    isolabel = isodata.isolabel()
    plt.plot(qvals,plotvals,**kwargs)
    mine = np.amin(plotvals)
    plt.xlim(np.array([0,1]))
    plt.ylim(np.array([0,5 * mine]))
    if type_ == 'pure':
        plt.xlabel('proportion of '+isolabel[isospike[0]]+' in '+isolabel[isospike[0]]+'-'+isolabel[isospike[1]]+' double spike')
    else:
        plt.xlabel('proportion of first rawspike in double spike')
    
    if errorratio is None:
        plt.ylabel(r'Error in $\alpha$ (1SD)')
    else:
        plt.ylabel('Error in '+isolabel[errorratio[0]]+'/'+isolabel[errorratio[1]]+' (1SD)')
    
    plt.title(isolabel[isoinv[0]]+', '+isolabel[isoinv[1]]+', '+isolabel[isoinv[2]]+', '+isolabel[isoinv[3]]+' inversion')


def errorcurveoptimalspike(isodata,type_ = 'pure',isospike = None,isoinv = None,errorratio = None,alpha = 0.0,beta = 0.0,plottype = 'default',**kwargs): 
    """Find the optimal double spike compositions and plot the corresponding error curves.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
                This is the only mandatory argument.
        type_ (str): type of spike, 'pure' or 'real'. Real spikes, such as those from
            Oak Ridge National Labs, contain impurities. See isodata.rawspike for 
            their assumed compositions. By default pure spikes are used.
        isospike (array): the isotopes used in the double spike e.g. [54, 57].
            By default all choices of 2 isotopes are tried.
        isoinv (array): the isotopes used in the inversion, e.g. [54, 56, 57, 58].
        errorratio (array): by default, the optimal spike is chosen as that which
            minimises the error on the natural fractionation factor (known as
            alpha). Instead, the optimiser can be told to minimise the
            error on a particular ratio by setting errorratio. e.g.
            setting errorratio=[58, 56] will minimise the error on 58Fe/56Fe.
        alpha, beta (float): there is a small dependance of the error on the fractionation
            factors (natural and instrumental). Values of alpha and
            beta can be set here if desired, although the effect on the optimal spikes
            is slight unless the fractionations are very large.
        plottype (str): by default, the error is plotted. By setting this to 'ppmperamu'
            an estimate of the ppm per amu is plotted instead.
        **kwargs: additional arguments are passed to the plot command
    
    Example:
        >>> errorcurveoptimalspike(IsoData('Fe'))
    
    See also optimalspike, errorcurve
    """
    if isoinv is None:
        if hasattr(isodata, 'isoinv'):
            isoinv = isodata.isoinv
        else:
            isoinv = isodata.isonum[0:4]
    
    # Convert isotope mass numbers to index numbers
    errorratio = isodata.isoindex(errorratio)
    isospike = isodata.isoindex(isospike)
    isoinv = isodata.isoindex(isoinv)
    # Find the optimal spikes
    os = optimalspike(isodata,type_,isospike,isoinv,errorratio,alpha,beta)
    isolabel = isodata.isolabel()
    rawspikelabel = isodata.rawspikelabel()
    for j in range(os['optspike'].shape[0]):
        if type_ == 'pure':
            spiked = np.where(os['optspike'][j,:] > 0)[0]
            leglabel = isolabel[spiked[0]]+'-'+isolabel[spiked[1]]
        else:
            spiked = np.where(os['optspikeprop'][j,:] > 0)[0]
            leglabel =rawspikelabel[spiked[0]]+'-'+rawspikelabel[spiked[1]]
        errorcurve(isodata,os['optspike'][j,:],isoinv,errorratio,alpha,beta,plottype,label=leglabel)
        
    if plottype=='ppmperamu':
        plt.ylim(np.array([0,5 * np.amin(os['optppmperamu'])]))
    else:
        plt.ylim(np.array([0,5 * np.amin(os['opterr'])]))
    
    plt.legend(loc='upper right')

if __name__=="__main__":
    from .isodata import IsoData
    idat = IsoData('Fe')    
    errorcurveoptimalspike(idat,'real')
    
    plt.show()
    



"""Routines for performing the double spike inversion."""

from .isodata import ratioproptorealprop, normalise_composition, ratio
import numpy as np
from scipy.optimize import fsolve 

def dsinversion(isodata, measured, spike=None, isoinv=None, standard=None):
    """Perform the double spike inversion for a given set of measurements.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
        measured (array): a matrix of beam intensities. Columns correspond to the
            different isotopes e.g. for Fe, first column is 54Fe, second is 56Fe,
            third is 57Fe, fourth is 58Fe. The matrix should have the same number
            of columns as there are isotopes available.
        spike (array): a composition vector for the spike. e.g. [0, 0, 0.5, 0.5] is a 50-50
            mix of 57Fe and 58Fe. If None this is read from isodata.
        isoinv (array): the four isotopes to use in the inversion, e.g [54, 56, 57, 58]. If
            None this is read from isodata.
        standard (array): standard composition or unspiked run data. If
            None this is read from isodata.
     
    Returns:
        This routine performs the double spike inversion on measured data to return the
        "true" composition of the sample. Output is returned as a dictionary with the
        following fields
            alpha: the inferred natural fractionations
            beta: the inferred instrumental fractionations
            prop: the inferred proportions of spike to sample
            sample: the inferred compositions of the sample
            mixture: the inferred compositions of the mixture
            
    Example:
        >>> dsinversion(IsoData('Fe'),measured,[0, 0, 0.5, 0.5],[54, 56, 57, 58])
    """
    # Get data from isodata if not supplied as arguments
    if spike is None:
        if isodata.spike is None:
            raise Exception("No spike given.")
        else:
            spike = isodata.spike
    if isoinv is None:
        if hasattr(isodata, 'isoinv'):
            isoinv = isodata.isoinv
        else:
            raise Exception("Inversion isotopes not specified.")
    if standard is None:
        standard = isodata.standard

    # Convert to numpy array if not already
    measured = np.array(measured)
    spike = np.array(spike)
    standard = np.array(standard)
    isoinv = np.array(isoinv)

    # Duplicate so all matrices same size
    nspike, nstandard, nmeasured = 1, 1, 1
    if spike.ndim >1:
        nspike = spike.shape[0]
    if measured.ndim >1:
        nmeasured = measured.shape[0]
    if standard.ndim >1:
        nstandard = standard.shape[0]
    nobs = max(nspike, nmeasured, nstandard)
    
    if spike.ndim==1:
        spike = np.tile(spike, (nobs,1))
    if measured.ndim==1:
        measured = np.tile(measured, (nobs,1))    
    if standard.ndim==1:
        standard = np.tile(standard, (nobs,1))  

    # Avoid division by zero errors for small values
    isoinv = isodata.isoindex(isoinv)
    if any(spike[0,isoinv] < 0.001):
        ix = np.argmax(spike[0,isoinv])
        deno = isoinv[ix]
        nume = isoinv[isoinv != deno]
        isoinv = np.concatenate((np.array([deno]),nume))

    # Take ratios based on the isotopes we are inverting
    P = np.log(ratio(isodata.mass, isoinv))
    n = ratio(standard, isoinv)
    T = ratio(spike, isoinv)
    m = ratio(measured,isoinv)
    
    z = np.zeros((nmeasured,3))
    for i in range(nobs):
        z[i,:] = dscorrection(P,n[i,:],T[i,:],m[i,:],xtol=1e-12)
    out = {}
    lambda_ = z[:,0]
    out['alpha'] = z[:,1]
    out['beta'] = z[:,2]
    
    isonum = np.arange(isodata.nisos())
    isonum = isonum[isonum != isoinv[0]]
    isonum = np.concatenate((np.array([isoinv[0]]), isonum))
    
    AP = np.log(ratio(isodata.mass, isonum))
    AT = ratio(spike, isonum)
    An = ratio(standard, isonum)
    Am = ratio(measured, isonum)
    
    # Calculate sample and mixture proportion, and proportion by mole
    AM = np.zeros_like(Am)
    AN = np.zeros_like(Am)
    out['prop']=np.zeros_like(out['alpha'])
    for i in np.arange(nobs):
        AM[i,:] = Am[i,:]*np.exp(- AP * out['beta'][i])
        AN[i,:] = An[i,:]*np.exp(- AP * out['alpha'][i])
        prop = ratioproptorealprop(lambda_[i], AT[i,:], AN[i,:])
        out['prop'][i] = prop
    
    out['sample'] = np.zeros_like(measured)
    out['mixture'] = np.zeros_like(measured)
    out['sample'][:,isonum[0]] = 1
    out['sample'][:,isonum[1:]] = AN
    out['mixture'][:,isonum[0]] = 1
    out['mixture'][:,isonum[1:]] = AM
    out['sample'] = normalise_composition(out['sample'])
    out['mixture'] = normalise_composition(out['mixture'])
    
    if nobs==1:
        # For single measurements make the output more compact
        out['alpha']=out['alpha'][0]
        out['beta']=out['beta'][0]
        out['prop']=out['prop'][0]
        out['sample']=np.squeeze(out['sample'])
        out['mixture']=np.squeeze(out['mixture'])
    
    return out

def dscorrection(P, n, T, m, **kwargs): 
    """Routine for double spike fractionation correction using isotope ratios as inputs.
    
    Args:
        P (array): log of ratio of atomic masses
        n (array): isotope ratios of standard/ unspiked run
        T (array): isotope ratios of spike
        m (array): isotope ratios of measured
        **kwargs: additional arguments passed to fsolve
          
    Returns:
        z (array): Spike ratio proportion (lambda),
                   natural fractionation (alpha),
                   and instrumental fractionation (beta)
                   as a vector z=(lambda, (1-lambda)*alpha, beta)
    """
    # start by solving the linear problem
    b = np.transpose((m - n))
    A = np.array([np.transpose((T - n)),np.transpose((np.multiply(- n,P))),np.transpose((np.multiply(m,P)))])
    y0 = np.linalg.solve(A,b)
    
    # by starting at the linear solution, solve the non-linear problem    
    y = fsolve(F, y0, args=(P,n,T,m), fprime=J, **kwargs)
    z = y
    z[1] = y[1] / (1 - y[0])
    return z

def F_params(y, P, n, T, m):
    """Determine main variables in the objective function."""
    lambda_ = y[0]
    alpha = y[1] / (1 - lambda_)
    beta = y[2]
    N = n * np.exp(- alpha * P)
    M = m * np.exp(- beta * P)
    return lambda_, alpha, beta, N, M
    
def F(y, P, n, T, m): 
    """The nonlinear equations to solve."""
    lambda_, alpha, beta, N, M = F_params(y, P, n, T, m)
    fval = lambda_*T + (1-lambda_)*N - M
    return fval

def J(y, P, n, T, m):
    """The Jacobian of the nonlinear equations -- can speed up root finding, but is not required."""
    lambda_, alpha, beta, N, M = F_params(y, P, n, T, m)
    dfdlambdaprime = T - N*(1 + alpha*P)
    dfdu = -N*P
    dfdbeta = M*P
    Jac = np.array([dfdlambdaprime,dfdu,dfdbeta]).T
    return Jac

if __name__=="__main__":
    from .isodata import IsoData
    isodata_fe = IsoData('Fe')
    isodata_fe.set_spike([0.0, 0.0, 0.5, 0.5])
    measured = np.array([0.2658, 4.4861, 2.6302, 2.6180])

    z = dsinversion(isodata_fe, measured)
    print(z)



"""Module for determining optimal double spikes."""

import numpy as np
import itertools
from scipy.special import binom
from scipy.optimize import minimize, LinearConstraint
from scipy.special import expit, logit
from .errors import errorestimate

def optimalspike(isodata,type_ = 'pure',isospike = None,isoinv = None,errorratio = None,alpha = 0.0,beta = 0.0): 
    """Find the optimal double spike composition and double spike-sample mixture proportions.
    
    Args:
        isodata: object of class IsoData, e.g. IsoData('Fe')
                This is the only mandatory argument.
        type(str): type of spike, 'pure' or 'real'. Real spikes, such as those from
            Oak Ridge National Labs, contain impurities. See isodata.rawspike
            for their assumed compositions. By default pure spikes are used.
        isospike (array): the isotopes used in the double spike e.g. [54, 57].
            By default all choices of 2 isotopes are tried.
        isoinv (array): the isotopes used in the inversion, e.g. [54, 56, 57, 58].
        errorratio (array): by default, the optimal double spike is chosen as that which
            minimises the error on the natural fractionation factor (known as
            alpha). Instead, the optimiser can be told to minimise the
            error on a particular ratio by setting errorratio. e.g.
            setting errorratio=[58, 56] will minimise the error on 58Fe/56Fe.
        alpha, beta (floats): there is a small dependance of the error on the fractionation
        factors (instrumental and natural, or alpha and beta). Values of alpha and
        beta can be set here if desired, although the effect on the optimal spikes
        is slight unless the fractionations are very large.

    Returns:
        All the outputs are provided as matrices. Each column represents an isotope
        (see isodata.isonum for the isotope numbers) e.g. for Fe the columns
        correspond to the isotopes 54Fe, 56Fe, 57Fe, 58Fe. The rows represent the
        different combinations of double spikes and isotopes being tried, in order of
        error: The first row is the best double spike, and the last row is the worst.
        optspike: the proportions of each isotope in the optimal double spike.
        optprop: the optimal proportion of spike in the double spike-sample mix.
        opterr: the error in the fractionation factor (or ratio if specified)
                for the optimal spike.
        optisoinv: the 4 isotopes used in the inversion.
        optspikeprop: the proportion of each raw spike in the optimal double spike.
        optppmperamu: an alternative expression of the error in terms of ppm per amu.

    Example:
        >>> isodata_fe = IsoData('Fe')
        >>> optspike,optprop,opterr,optisoinv,optspikeprop,optppmperamu = optimalspike(isodata_fe,'pure')
    """
    # Convert isotope mass numbers to index numbers
    errorratio = isodata.isoindex(errorratio)
    isospike = isodata.isoindex(isospike)
    isoinv = isodata.isoindex(isoinv)
    
    # If don't specify inversion isotopes, do all possible combinations
    if isoinv is None:
        isoinv = list(itertools.combinations(np.arange(isodata.nisos()), 4))
    else:
        isoinv = list([np.array(isoinv)])
    
    # Work out all combinations of inversion isotopes and spiking isotopes
    isoinvvals = []
    isospikevals = []
    for i in range(len(isoinv)):
        if isospike is None:
            if type_ == 'pure':
                # look at all combinations of spikes from the inversion isotopes
                isospikev = list(itertools.combinations(isoinv[i], 2))
            else:
                if isodata.nrawspikes() == 0:
                    return {} # can't proceed if no single spikes to use
                # look at all combinations of spikes from the all rawspikes
                isospikev = list(itertools.combinations(np.arange(isodata.nrawspikes()), 2))
        else:
            isospikev = list([isospike])
            #if len(set(isospike).intersection(set(isoinv[i]))) == 2:
                #isospikev = list([isospike])
            #else:
                #isospikev = None
        
        if isospikev is not None:
            isospikevals.append(isospikev)
            isoinvvals.append(np.tile(isoinv[i],(len(isospikev),1)))
    isoinvvals = np.vstack(isoinvvals)
    isospikevals = np.vstack(isospikevals)
   
    optspikes = []
    optprops = []
    opterrs = []
    optppmperamus = []
    optspikeprops = []
    
    for i in range(len(isoinvvals)):
        try:
            optspike,optprop,opterr,optspikeprop,optppmperamu = singleoptimalspike(isodata,type_,isospikevals[i,:],isoinvvals[i,:],errorratio,alpha,beta)
        except:
            # try to fail gracefully
            optspike = np.zeros(isodata.nisos())
            optprop = 0.0
            opterr = 1e32
            optspikeprop = np.zeros(isodata.nisos())
            optppmperamu = 1e32
                
        optspikes.append(optspike)
        optprops.append(optprop)
        opterrs.append(opterr)
        optppmperamus.append(optppmperamu)
        optspikeprops.append(optspikeprop)
    
    optspike = np.vstack(optspikes)
    optspikeprop = np.vstack(optspikeprops)
    optprop = np.array(optprops)
    opterr = np.array(opterrs)
    optppmperamu = np.array(optppmperamus)
    optisoinv = isoinvvals
    
    ## Sort in ascending order of error
    ix = np.argsort(opterr)
    
    # avoid masses of output by limiting to all possibilites in case of pure spikes
    max_noutput = min(len(ix), int(6*binom(isodata.nisos(),4)))
    ix = ix[0:max_noutput]
    
    out = {'optspike': optspike[ix,:],
           'optprop': optprop[ix],
           'opterr': opterr[ix],
           'optisoinv': isodata.isonum[optisoinv[ix,:]],
           'optspikeprop': optspikeprop[ix,:],
           'optppmperamu': optppmperamu[ix]}
           
    return out
    
def singleoptimalspike(isodata, type_ = 'pure',isospike = None,isoinv = None,errorratio = None, alpha = 0.0, beta = 0.0): 
    """Calculate the composition of the optimal double spike given the isotopes used in the inversion and of those the isotopes we are spiking."""
    if type_ == "pure":
        spikevector1 = np.zeros(isodata.nisos())
        spikevector1[isospike[0]] = 1.0
        spikevector2 = np.zeros(isodata.nisos())
        spikevector2[isospike[1]] = 1.0
    else:
        spikevector1 = isodata.rawspike[isospike[0], :]
        spikevector2 = isodata.rawspike[isospike[1], :]
    
    # Helpful to rescale the error, to make everything roughly order 1 for the optimiser
    initialerror, _ = errorestimate(isodata,0.5,0.5*spikevector1 + (1 - 0.5)*spikevector2,isoinv,errorratio,beta,alpha)

    def objective(y):
        p = expit(y[0])  # use expit transformation to keep things in range [0,1]
        q = expit(y[1])
        
        error, ppmperamu = errorestimate(isodata,p,q*spikevector1 + (1.0 - q)*spikevector2,isoinv,errorratio,beta,alpha)
        return error/initialerror 
    
    y0 = np.array([0.0, 0.0])
    res = minimize(objective, y0, tol = 1e-9)
    
    y = res.x
    p = expit(y[0])
    q = expit(y[1])
    
    optprop = p
    optspike = q * spikevector1 + (1 - q) * spikevector2
    opterr,optppmperamu = errorestimate(isodata,p,q*spikevector1 + (1 - q)*spikevector2,isoinv,errorratio,beta,alpha)
    
    optspikeprop=np.zeros_like(optspike)
    optspikeprop[isospike[0]]=q
    optspikeprop[isospike[1]]=1-q
    
    return optspike,optprop,opterr,optspikeprop,optppmperamu

if __name__=="__main__":
    from .isodata import IsoData
    ##isodata = IsoData('Fe')
    isodata_ca = IsoData('Ca')
    isoinv = [40, 42, 44, 48]
    
    isospike = [2,5]
    
    print(optimalspike(isodata_ca,'real', isoinv = isoinv, isospike = isospike))
